@@ -56,7 +56,6 @@
 #include "wimlib/alloca.h"
 #include "wimlib/assert.h"
 #include "wimlib/blob_table.h"
-#include "wimlib/capture.h"
 #include "wimlib/dentry.h"
 #include "wimlib/encoding.h"
 #include "wimlib/endianness.h"
@@ -64,6 +63,8 @@
 #include "wimlib/metadata.h"
 #include "wimlib/paths.h"
 #include "wimlib/progress.h"
+#include "wimlib/scan.h"
+#include "wimlib/test_support.h"
 #include "wimlib/xml_windows.h"
 
 /* Saved specification of a "primitive" update operation that was performed.  */
@@ -786,9 +787,9 @@ execute_add_command(struct update_command_journal *j,
 	tchar *fs_source_path;
 	tchar *wim_target_path;
 	const tchar *config_file;
-	struct capture_params params;
+	struct scan_params params;
 	struct capture_config config;
-	capture_tree_t capture_tree = platform_default_capture_tree;
+	scan_tree_t scan_tree = platform_default_scan_tree;
 	struct wim_dentry *branch;
 
 	add_flags = add_cmd->add.add_flags;
@@ -800,7 +801,12 @@ execute_add_command(struct update_command_journal *j,
 
 #ifdef WITH_NTFS_3G
 	if (add_flags & WIMLIB_ADD_FLAG_NTFS)
-		capture_tree = ntfs_3g_build_dentry_tree;
+		scan_tree = ntfs_3g_build_dentry_tree;
+#endif
+
+#ifdef ENABLE_TEST_SUPPORT
+	if (add_flags & WIMLIB_ADD_FLAG_GENERATE_TEST_DATA)
+		scan_tree = generate_dentry_tree;
 #endif
 
 	ret = get_capture_config(config_file, &config,
@@ -826,7 +832,7 @@ execute_add_command(struct update_command_journal *j,
 
 	if (WIMLIB_IS_WIM_ROOT_PATH(wim_target_path))
 		params.add_flags |= WIMLIB_ADD_FLAG_ROOT;
-	ret = (*capture_tree)(&branch, fs_source_path, &params);
+	ret = (*scan_tree)(&branch, fs_source_path, &params);
 	if (ret)
 		goto out_destroy_config;
 
@@ -861,7 +867,7 @@ execute_add_command(struct update_command_journal *j,
 		/* If a capture configuration file was explicitly specified when
 		 * capturing an image in WIMBoot mode, save it as
 		 * /Windows/System32/WimBootCompress.ini in the WIM image. */
-		ret = platform_default_capture_tree(&branch, config_file, &params);
+		ret = platform_default_scan_tree(&branch, config_file, &params);
 		if (ret)
 			goto out_destroy_config;
 
@@ -1206,6 +1212,9 @@ check_add_command(struct wimlib_update_command *cmd,
 			  WIMLIB_ADD_FLAG_NO_REPLACE |
 			  WIMLIB_ADD_FLAG_TEST_FILE_EXCLUSION |
 			  WIMLIB_ADD_FLAG_SNAPSHOT |
+		#ifdef ENABLE_TEST_SUPPORT
+			  WIMLIB_ADD_FLAG_GENERATE_TEST_DATA |
+		#endif
 			  WIMLIB_ADD_FLAG_FILE_PATHS_UNNEEDED))
 		return WIMLIB_ERR_INVALID_PARAM;
 
@@ -1213,7 +1222,7 @@ check_add_command(struct wimlib_update_command *cmd,
 
 #ifndef WITH_NTFS_3G
 	if (add_flags & WIMLIB_ADD_FLAG_NTFS) {
-		ERROR("NTFS-3g capture mode is unsupported because wimlib "
+		ERROR("NTFS-3G capture mode is unsupported because wimlib "
 		      "was compiled --without-ntfs-3g");
 		return WIMLIB_ERR_UNSUPPORTED;
 	}
